#include "stdafx.h"
// dump.cpp -- dump out ID3DX mesh information
//
// An ID3DXMesh contains a group of independent triangles, stored in the
// associated index and vertex buffers.  Drawsubset() amounts to
// a DrawIndexedPrimitive with D3DPT_TRIANGLELIST and the index and vertex
// data set on the device.  The mesh is said to contain a certain number
// of triangular 'faces', which are composed of vertices.  In a typical
// model there is quite a bit of vertex sharing among the triangles in
// the mesh.  For instance:
//
//      A---B---C---D---E    The points B, C, D, G, H, and I are all
// +Y   |0 /|2 /|4 /|6 /|    shared among two triangles, provided that
//  ^   | / | / | / | / |    adjacent triangles share vertex normals and
//  |   |/ 1|/ 3|/ 5|/ 7|    other per-vertex data.  This is typically the
//  |   F---G---H---I---J    case.
//  | 
//  X ----> +X
//
// Each triangle in the mesh is indexed from 0 through N-1 for N faces.
// The mesh above has 8 triangles, indexed from 0 through 7.  The
// triangles are formed from ordered triplets of vertices.  If the
// triangles were independent 24 vertices would be required to describe
// them, but because of the sharing we need only 10 vertices.
//
// The index buffer entries i*3, i*3+1 and i*3+2 contain the vertex
// buffer indices for triangle i.  ID3DXMesh supports both 16-bit and
// 32-bit vertex buffer indices.  The GetOptions method on ID3DXBaseMesh
// can be used to determine if the indices are 16 or 32-bits.
//
// When you create a mesh with one of the D3DX functions, you get an
// adjacency array.  The adjacency array is an array of integers that
// describe which other triangles are adjacent to any given triangle.
// The array entries i*3, i*3+1 and i*3+2 give the indices of the
// triangles that are adjacent to triangle i.  If a triangle has an
// edge that is at the boundary of the mesh, such as the segments AB,
// BC, CD, DE, EJ, AF, FG, GH, HI and IJ above, then the adjacency
// array entry for that edge of the triangle is -1.
//
// For triangle ABF, the adjacency array entries i*3, i*3+1 and i*3+2
// contain the triangle indices for the triangles adjacent to sides AB,
// BF, and FA, respectively.  Sides AB and FA are at the edge of the
// mesh, so their adjacency entries are -1.  Side BF is adjacent to
// triangle 1, so its adjacency entry is 1.
//
// Here is a dump of cube.x which comes with the SDK:
//
// File: "D:\dx8sdk\samples\Multimedia\Media\cube.x" 12 faces, 8 vertices, XYZ NORMAL T[0]:rs
// 0: STREAM(false, 0)
// 1: STREAMDATA(3D float, 0)
// 2: STREAMDATA(3D float, 3)
// 3: STREAMDATA(2D float, 7)
// 4: END
// Adjacency: 12 faces.
// 0: [3, 5, 1], [0, 7, 9], [9, 11, 3]
// 3: [2, 4, 0], [3, 11, 5], [4, 6, 0]
// 6: [5, 10, 7], [6, 8, 1], [7, 10, 9]
// 9: [8, 2, 1], [8, 6, 11], [10, 4, 2]
// Materials:
// 0: D=RGBA<1,1,1,1>, A=RGBA<0,0,0,-0.00132704>, S=RGBA<0,0,0,1>, E=RGBA<0,0,0,1>, p=0
// Face Indices: 12 faces, 36 indices.
// 0: [0, 1, 2], [0, 2, 3], [0, 4, 1]
// 3: [0, 5, 1], [1, 5, 6], [1, 6, 7]
// 6: [2, 6, 7], [2, 7, 3], [3, 7, 0]
// 9: [3, 4, 0], [4, 7, 6], [4, 6, 5]
// Vertices: 8
// 0: { (0.996094, 1, -0.99585), <0.333333, 0.666667, -0.666667>, (0, 0) }
// 1: { (-0.99585, 1, -0.99585), <-0.816497, 0.408248, -0.408248>, (0, 0) }
// 2: { (-0.99585, 1, 0.996094), <-0.333333, 0.666667, 0.666667>, (0, 0) }
// 3: { (0.996094, 1, 0.996094), <0.816497, 0.408248, 0.408248>, (0, 0) }
// 4: { (0.996094, -0.999756, -0.99585), <0.666667, -0.666667, -0.333333>, (0, 0) }
// 5: { (-0.99585, -0.999756, -0.99585), <-0.408248, -0.408248, -0.816497>, (0, 0) }
// 6: { (-0.99585, -0.999756, 0.996094), <-0.666667, -0.666667, 0.333333>, (0, 0) }
// 7: { (0.996094, -0.999756, 0.996094), <0.408248, -0.408248, 0.816497>, (0, 0) }
//
// Notice that no adjacency values are -1 because a cube is entirely
// closed and we only see the outside of the cube.  The first triangle
// has vertices (rounding to integers): (1, 1, -1), (-1, 1, -1), (-1, 1, 1)
// which is in the xz plane since y is constant.
//
// About the code:
//
// There's not much documentation through this code because how much
// documentation does a print statement need?  However, the basic idea
// here is to leverage C++ templates and the function overloading in
// operator<< to factor out formatting tasks as much as possible and then
// compose the factored results as needed for the various structures.
// Beyond that are some little routines for getting a string representation
// of an FVF.  This isn't useful in real applications, only something for
// debugging and programming.  So while there's lots of code here it isn't
// something the typical Direct3D application is going to be doing and
// its not terribly complicated.
//
// See "The C++ Standard Library", by Nicolai Josuttis, for information
// on all the operator<< stuff.
//
// Copyright (C) 2000-2001, Rich Thomson, all rights reserved.
//////////////////////////////////////////////////////////////////////

// Build up a cascade of structs to drive operator<< formatting.
struct s_tex_coords2
{
    float u, v;
};

//////////////////////////////////////////////////////////////////////
struct s_pos3
{
    float x, y, z;
};

//////////////////////////////////////////////////////////////////////
struct s_vec3
{
    float i, j, k;
};

//////////////////////////////////////////////////////////////////////
struct s_vertex
{
    s_pos3 m_pos;
    s_vec3 m_normal;
    s_tex_coords2 m_tex0;
};

//////////////////////////////////////////////////////////////////////
struct s_vertex_xyz
{
    s_pos3 m_pos;
};

//////////////////////////////////////////////////////////////////////
struct s_vertex_xyz_normal
{
    s_pos3 m_pos;
    s_vec3 m_normal;
};

//////////////////////////////////////////////////////////////////////
struct s_adjacency
{
    s_adjacency(const int *adj) : m_adj(adj) {}
    const int *m_adj;
};

//////////////////////////////////////////////////////////////////////
struct named_dword {
    DWORD m_value;
    LPCTSTR m_name;
};

//////////////////////////////////////////////////////////////////////
template <typename T>
class c_index_lock
{
public:
    c_index_lock(IDirect3DIndexBuffer8 *buffer,
                 DWORD flags = D3DLOCK_READONLY,
                 UINT start = 0, UINT size = 0)
                 : m_buffer(buffer)
    {
        THR(buffer->Lock(start, size, &m_data, flags));
    }
    ~c_index_lock()
    {
        // can't use THR() here because destructors shouldn't throw
        const HRESULT hr = m_buffer->Unlock();
        _ASSERT(SUCCEEDED(hr));
    }
    const T *data() const
    {
        return static_cast<T *>(static_cast<void *>(m_data));
    }
    T *data()
    {
        return static_cast<T *>(static_cast<void *>(m_data));
    }

private:
    CComPtr<IDirect3DIndexBuffer8> m_buffer;
    BYTE *m_data;
};

//////////////////////////////////////////////////////////////////////
template <typename T>
class c_vertex_lock
{
public:
    c_vertex_lock(IDirect3DVertexBuffer8 *buffer,
                  DWORD flags = D3DLOCK_READONLY,
                  UINT start = 0, UINT size = 0)
                  : m_buffer(buffer)
    {
        THR(buffer->Lock(start, size, &m_data, flags));
    }
    ~c_vertex_lock()
    {
        const HRESULT hr = m_buffer->Unlock();
        _ASSERT(SUCCEEDED(hr));
    }
    const T *data() const
    {
        return static_cast<T *>(static_cast<void *>(m_data));
    }
    T *data()
    {
        return static_cast<T *>(static_cast<void *>(m_data));
    }

private:
    CComPtr<IDirect3DVertexBuffer8> m_buffer;
    BYTE *m_data;
};
                  
//////////////////////////////////////////////////////////////////////
inline rt::tostream
operator<<(rt::tostream &str, const s_tex_coords2 &coords)
{
    return str << _T("(") << coords.u << _T(", ") << coords.v << _T(")");
}

//////////////////////////////////////////////////////////////////////
inline rt::tostream
operator<<(rt::tostream &str, const s_pos3 &pos)
{
    return str << _T("(") << pos.x << _T(", ") << pos.y << _T(", ")
        << pos.z << _T(")");
}

//////////////////////////////////////////////////////////////////////
inline rt::tostream
operator<<(rt::tostream &str, const s_vec3 &vec)
{
    return str << _T("<") << vec.i << _T(", ") << vec.j << _T(", ")
        << vec.k << _T(">");
}

//////////////////////////////////////////////////////////////////////
inline rt::tostream
operator<<(rt::tostream &str, const s_vertex &vertex)
{
    return str << _T("{ ") << vertex.m_pos << _T(", ")
        << vertex.m_normal << _T(", ")
        << vertex.m_tex0 << _T(" }");
}

//////////////////////////////////////////////////////////////////////
inline rt::tostream
operator <<(rt::tostream &str, const s_vertex_xyz &vertex)
{
    return str << _T("{ ") << vertex.m_pos << _T(" }");
}

//////////////////////////////////////////////////////////////////////
inline rt::tostream
operator<<(rt::tostream &str, const s_vertex_xyz_normal &vertex)
{
    return str << _T("{ ") << vertex.m_pos << _T(", ")
        << vertex.m_normal << _T(" }");
}

//////////////////////////////////////////////////////////////////////
inline rt::tostream
operator<<(rt::tostream &str, const s_adjacency &adj)
{
    return str << _T("[")
        << adj.m_adj[0] << _T(", ")
        << adj.m_adj[1] << _T(", ")
        << adj.m_adj[2] << _T("]");
}

//////////////////////////////////////////////////////////////////////
inline rt::tostream
operator<<(rt::tostream &str, const D3DCOLORVALUE &col)
{
    return str << _T("RGBA<")
        << col.r << _T(",")
        << col.g << _T(",")
        << col.b << _T(",")
        << col.a << _T(">");
}

//////////////////////////////////////////////////////////////////////
inline rt::tostream
operator<<(rt::tostream &str, const D3DMATERIAL8 &mat)
{
    str << _T("D=") << mat.Diffuse;
    str << _T(", A=") << mat.Ambient;
    str << _T(", S=") << mat.Specular;
    str << _T(", E=") << mat.Emissive;
    str << _T(", p=") << mat.Power;
    return str;
}

//////////////////////////////////////////////////////////////////////
inline rt::tostream
operator<<(rt::tostream &str, const D3DXMATERIAL &mat)
{
    if (mat.pTextureFilename)
    {
        USES_CONVERSION;
        str << _T("\"") << A2CT(mat.pTextureFilename) << _T("\" ");
    }
    return str << mat.MatD3D;
}

//////////////////////////////////////////////////////////////////////
inline void
ODS(rt::tostringstream &str)
{
    str << std::endl << std::ends;
    ::OutputDebugString(str.str().c_str());
    str.seekp(0);
}

//////////////////////////////////////////////////////////////////////
void
dump_adjacency(const DWORD *tmp, UINT start, UINT count)
{
    const INT *adj = reinterpret_cast<const INT *>(tmp);
    rt::tostringstream msg;
    const UINT end = start + count;
    for (UINT i = start; i+3 < end; i += 3)
    {
        msg << i << _T(": ")
            << s_adjacency(adj + (i + 0)*3) << _T(", ")
            << s_adjacency(adj + (i + 1)*3) << _T(", ")
            << s_adjacency(adj + (i + 2)*3);
        ODS(msg);
    }
    if (i < end)
    {
        msg << i << _T(": ");
        while (i < end)
        {
            msg << s_adjacency(adj + i*3);
            if (++i < end)
            {
                msg << _T(", ");
            }
        }
        ODS(msg);
    }
}

//////////////////////////////////////////////////////////////////////
void
dump_adjacency(UINT num_faces, ID3DXBuffer *buffer)
{
    const DWORD *adjacency =
        static_cast<DWORD *>(buffer->GetBufferPointer());
    rt::tostringstream msg;
    msg << _T("Adjacency: ") << num_faces << _T(" faces.");
    ODS(msg);
    if (num_faces < 25)
    {
        dump_adjacency(adjacency, 0, num_faces);
    }
    else
    {
        dump_adjacency(adjacency, 0, 12);
        msg << _T("...");
        ODS(msg);
        dump_adjacency(adjacency, num_faces-12, 12);
    }
}

//////////////////////////////////////////////////////////////////////
void
dump_materials(DWORD num_materials, ID3DXBuffer *buffer)
{
    rt::tostringstream msg;
    msg << _T("Materials:");
    ODS(msg);
    const D3DXMATERIAL *materials =
        static_cast<D3DXMATERIAL *>(buffer->GetBufferPointer());
    for (int i = 0; i < num_materials; i++)
    {
        msg << i << _T(": ") << materials[i];
        ODS(msg);
    }
}

//////////////////////////////////////////////////////////////////////
void
dump_subsets(ID3DXMesh *mesh)
{
    DWORD num_subsets = 0;
    THR(mesh->GetAttributeTable(NULL, &num_subsets));
    if (num_subsets)
    {
        std::vector<D3DXATTRIBUTERANGE> subsets(num_subsets);
        THR(mesh->GetAttributeTable(&subsets[0], &num_subsets));
        rt::tostringstream msg;
        for (int i = 0; i < num_subsets; i++)
        {
            msg << i << _T(": ID=") << subsets[i].AttribId << _T(", Faces=")
                << subsets[i].FaceStart << _T("+") << subsets[i].FaceCount
                << _T(", Vertices=") << subsets[i].VertexCount << _T("+")
                << subsets[i].VertexCount;
            ODS(msg);
        }
    }
}

//////////////////////////////////////////////////////////////////////
template <typename T>
void
dump_indices(UINT start, UINT count, const T *indices)
{
    rt::tostringstream buff;
    const UINT end = start + count;
    for (UINT i = start; i+3 < end; i += 3)
    {
        buff << i << _T(": [")
            << indices[(i+0)*3+0] << _T(", ")
            << indices[(i+0)*3+1] << _T(", ")
            << indices[(i+0)*3+2] << _T("], [")
            << indices[(i+1)*3+0] << _T(", ")
            << indices[(i+1)*3+1] << _T(", ")
            << indices[(i+1)*3+2] << _T("], [")
            << indices[(i+2)*3+0] << _T(", ")
            << indices[(i+2)*3+1] << _T(", ")
            << indices[(i+3)*3+2] << _T("]");
        ODS(buff);
    }
    if (i < end)
    {
        buff << i << _T(": ");
        while (i < end)
        {
            buff << _T("[")
                << indices[i*3+0] << _T(", ")
                << indices[i*3+1] << _T(", ")
                << indices[i*3+2] << _T("]");
            if (++i < end)
            {
                buff << _T(", ");
            }
        }
        ODS(buff);
    }
}

//////////////////////////////////////////////////////////////////////
//
// This one -does- need some explaining because of the funky
// T * = 0 function argument.  It turns out that saying
// dump_vertices<T>(mesh) in the function call wasn't enough to resolve
// out the proper function.  So I guessed that it was having a hard
// time matching the template because T didn't appear in the argument
// list.  So I added a bogus T * parameter to the end with a default
// value of 0.  The argument itself is -only- there to appease the
// quirks and non-compliance of VC++6 in terms of template support.
// I hear VC++7 (i.e. VS.NET) is better, but haven't tried example out
// yet.
//
template <typename T>
dump_indices(ID3DXMesh *mesh, T * = 0)
{
    rt::tostringstream buff;
    UINT num_faces = mesh->GetNumFaces();
    buff << _T("Face Indices: ") << num_faces << _T(" faces, ")
        << (num_faces*3) << _T(" indices.");
    ODS(buff);

    CComPtr<IDirect3DIndexBuffer8> ib;
    THR(mesh->GetIndexBuffer(&ib));
    c_index_lock<T> indices(ib);

    if (num_faces < 25)
    {
        dump_indices<T>(0, num_faces, indices.data());
    }
    else
    {
        dump_indices<T>(0, 12, indices.data());
        buff << _T("...");
        ODS(buff);
        dump_indices<T>(num_faces-12, 12, indices.data());
    }
}

//////////////////////////////////////////////////////////////////////
template <typename T>
void
dump_vertices(UINT start, UINT count, const T *vertices)
{
    rt::tostringstream buff;
    const UINT end = start + count;
    for (UINT i = start; i < end; i++)
    {
        buff << i << _T(": ") << vertices[i];
        ODS(buff);
    }
}

//////////////////////////////////////////////////////////////////////
//
// This one -does- need some explaining because of the funky
// T * = 0 function argument.  It turns out that saying
// dump_vertices<T>(mesh) in the function call wasn't enough to resolve
// out the proper function.  So I guessed that it was having a hard
// time matching the template because T didn't appear in the argument
// list.  So I added a bogus T * parameter to the end with a default
// value of 0.  The argument itself is -only- there to appease the
// quirks and non-compliance of VC++6 in terms of template support.
// I hear VC++7 (i.e. VS.NET) is better, but haven't tried example out
// yet.
//
template <typename T>
void
dump_vertices(ID3DXMesh *mesh, T * = 0)
{
    const UINT num_vertices = mesh->GetNumVertices();
    rt::tostringstream buff;
    buff << _T("Vertices: ") << num_vertices;
    ODS(buff);

    CComPtr<IDirect3DVertexBuffer8> vb;
    THR(mesh->GetVertexBuffer(&vb));
    c_vertex_lock<T> vertices(vb);
    
    if (num_vertices < 25)
    {
        dump_vertices<T>(0, num_vertices, vertices.data());
    }
    else
    {
        dump_vertices<T>(0, 12, vertices.data());
        buff << _T("...");
        ODS(buff);
        dump_vertices<T>(num_vertices-12, 12, vertices.data());
    }
}

//////////////////////////////////////////////////////////////////////
void
dump_geometry(ID3DXMesh *mesh)
{
    dump_subsets(mesh);

    if (D3DXMESH_32BIT & mesh->GetOptions())
    {
        dump_indices<DWORD>(mesh);
    }
    else
    {
        dump_indices<WORD>(mesh);
    }
    const DWORD fvf = mesh->GetFVF();
    const DWORD XYZ = D3DFVF_XYZ;
    const DWORD NORMAL = D3DFVF_NORMAL;
    const DWORD TEX1 = D3DFVF_TEX1;
    const DWORD SIZE2 = D3DFVF_TEXCOORDSIZE2(0);
    const DWORD combo = XYZ | NORMAL | TEX1 | SIZE2;
    switch (fvf)
    {
    case D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE2(0):
        dump_vertices<s_vertex>(mesh);
        break;

    case D3DFVF_XYZ | D3DFVF_NORMAL:
        dump_vertices<s_vertex_xyz_normal>(mesh);
        break;

    case D3DFVF_XYZ:
        dump_vertices<s_vertex_xyz>(mesh);
        break;

    default:
        _ASSERT(0);
        break;
    }
}

//////////////////////////////////////////////////////////////////////
inline rt::tstring
fvf(DWORD fvf)
{
    rt::tstring text;
    switch (fvf & D3DFVF_POSITION_MASK)
    {
    case D3DFVF_XYZ: text = _T("XYZ"); break;
    case D3DFVF_XYZB1: text = _T("XYZB1"); break;
    case D3DFVF_XYZB2: text = _T("XYZB2"); break;
    case D3DFVF_XYZB3: text = _T("XYZB3"); break;
    case D3DFVF_XYZB4: text = _T("XYZB4"); break;
    case D3DFVF_XYZB5: text = _T("XYZB5"); break;
    default: text = _T("Unknown position mask value"); break;
    }

    const struct named_dword fvf_names[] =
    {
        D3DFVF_RESERVED0, _T("RESERVED0"),
        D3DFVF_NORMAL, _T("NORMAL"),
        D3DFVF_PSIZE, _T("PSIZE"),
        D3DFVF_DIFFUSE, _T("DIFFUSE"),
        D3DFVF_SPECULAR, _T("SPECULAR"),
        D3DFVF_LASTBETA_UBYTE4, _T("LASTBETA_UBYTE4"),
        D3DFVF_RESERVED2, _T("RESERVED2")
    };
    int i;
    for (i = 0; i < NUM_OF(fvf_names); i++)
    {
        if (fvf & fvf_names[i].m_value)
        {
            if (text.length())
            {
                text += _T(" ");
            }
            text += fvf_names[i].m_name;
        }
    }
    const UINT num_tex_coords =
        (fvf & D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT;
    fvf >>= 16;
    rt::tstring stages(_T("01234567"));
    for (i = 0; i < num_tex_coords; i++)
    {
        text += _T(" T[") + stages.substr(i, 1) + _T("]:");
        switch (fvf & 0x3)
        {
        case D3DFVF_TEXTUREFORMAT2:          // Two floating point values
            text += _T("rs");
            break;

        case D3DFVF_TEXTUREFORMAT1:          // One floating point value
            text += _T("r");
            break;

        case D3DFVF_TEXTUREFORMAT3:          // Three floating point values
            text += _T("rst");
            break;

        case D3DFVF_TEXTUREFORMAT4:          // Four floating point values
            text += _T("rstu");
            break;
        }
        fvf >>= 2;
    }

    return text;
}

//////////////////////////////////////////////////////////////////////
inline bool
token_data_load_type(DWORD token)
{
    return (token & D3DVSD_DATALOADTYPEMASK) ? true : false;
}

//////////////////////////////////////////////////////////////////////
#define DECL_FIELD(fn_name_, bit_name_) \
inline DWORD \
token_##fn_name_(DWORD token) \
{ \
    return (token & D3DVSD_##bit_name_##MASK) >> D3DVSD_##bit_name_##SHIFT; \
}
DECL_FIELD(type, TOKENTYPE)
DECL_FIELD(data_type, DATATYPE)
DECL_FIELD(skip_count, SKIPCOUNT)
DECL_FIELD(vertex_reg, VERTEXREG)
DECL_FIELD(const_count, CONSTCOUNT)
DECL_FIELD(const_rs, CONSTRS)
DECL_FIELD(ext_count, EXTCOUNT)
DECL_FIELD(ext_info, EXTINFO)
#undef DECL_FIELD

//////////////////////////////////////////////////////////////////////
inline LPCTSTR
token_data_type_name(DWORD token)
{
    static LPCTSTR type_names[] =
    {
        _T("1D float"),
        _T("2D float"),
        _T("3D float"),
        _T("4D float"),
        _T("D3D Color"),
        _T("4D unsigned byte"),
        _T("2D signed short"),
        _T("4D signed short")
    };
    return type_names[token_data_type(token)];
}

//////////////////////////////////////////////////////////////////////
rt::tstring
token(DWORD token)
{

    rt::tostringstream str;
    switch (token_type(token))
    {
    case D3DVSD_TOKEN_NOP:
        str << _T("NOP");
        break;

    case D3DVSD_TOKEN_STREAM:
        str << _T("STREAM(") << std::boolalpha << token_data_load_type(token)
            << _T(", ") << token_vertex_reg(token) << _T(")");
        break;

    case D3DVSD_TOKEN_STREAMDATA:
        str << _T("STREAMDATA(");
        if (token_data_load_type(token))
        {
            str << _T("skip=") << token_data_type_name(token);
        }
        else
        {
            str << token_data_type_name(token) << _T(", ")
                << token_vertex_reg(token);
        }
        str << _T(")");
        break;

    case D3DVSD_TOKEN_TESSELLATOR:
        str << _T("TESSELATOR(")
            << (token_data_load_type(token) ? _T("normals, ") : _T("u,v, "))
            << token_data_type_name(token)
            << _T(", ") << token_vertex_reg(token)
            << _T(")");
        break;

    case D3DVSD_TOKEN_CONSTMEM:
        str << _T("CONSTMEM(count=")
            << token_const_count(token)
            << _T(", address=") << token_const_rs(token)
            << _T(")");
        break;

    case D3DVSD_TOKEN_EXT:
        str << _T("EXT");
        break;

    case D3DVSD_TOKEN_END:
        str << _T("END");
        break;

    default:
        str << _T("Unknown token: ") << std::hex << std::setw(8) << token;
    }
    return str.str();
}

//////////////////////////////////////////////////////////////////////
void
dump_file_info(LPCTSTR mesh_file, ID3DXMesh *mesh)
{
    rt::tostringstream msg;
    msg << _T("File: \"") << mesh_file << _T("\" ") << mesh->GetNumFaces()
        << _T(" faces, ") << mesh->GetNumVertices() << _T(" vertices, ")
        << fvf(mesh->GetFVF());
    ODS(msg);

    DWORD decl[MAX_FVF_DECL_SIZE];
    THR(mesh->GetDeclaration(decl));
    for (UINT i = 0; decl[i] != ~0; i++)
    {
        msg << i << _T(": ") << token(decl[i]);
        ODS(msg);
    }
    msg << i << _T(": END");
    ODS(msg);
}

//////////////////////////////////////////////////////////////////////
void
dump_mesh(IDirect3DDevice8 *device, LPCTSTR mesh_file)
{
    DWORD num_materials = 0;
    CComPtr<ID3DXBuffer> adjacency;
    CComPtr<ID3DXBuffer> materials;
    CComPtr<ID3DXMesh> mesh;
    USES_CONVERSION;
    THR(D3DXLoadMeshFromX(T2A(const_cast<LPTSTR>(mesh_file)),
                          D3DXMESH_SYSTEMMEM, device, &adjacency,
                          &materials, &num_materials, &mesh));

    // dump global mesh info
    dump_file_info(mesh_file, mesh);

    // dump materials and adjacency
    dump_adjacency(mesh->GetNumFaces(), adjacency);
    dump_materials(num_materials, materials);

    // dump vertex, index and subset data
    dump_geometry(mesh);
}
